import sys
from pathlib import Path

DATA_TYPES = {
    "csv",
    "ipc",
    "parquet",
    "database",
    "json",
    "ndjson",
    "avro",
    "excel",
    "delta",
}


def activate(
    df,
    fetch: int = None,
    streaming: bool = None,
):
    df = (
        df.fetch(n_rows=fetch, streaming=streaming)
        if fetch
        else df.collect(streaming=streaming)
    )
    return df


def polars(
    *paths: str,
    read_type: str = None,
    streaming: str = None,  # actually bool
    fetch: int = None,
    join: str = None,
    on: str = None,
    left_on: str = None,
    right_on: str = None,
    suffix: str = "_right",
    validate: str = "m:m",
    head: int = None,
    tail: int = None,
    sample: int = None,
    method: str = None,
    write_path: str = None,
    **kwargs,
):
    """Read and transform tabular files using polars."""

    """Workaround for unexpected behavior of Fire"""
    kwargs.pop("streaming", None)
    kwargs.pop("fetch", None)
    kwargs.pop("join", None)
    kwargs.pop("on", None)
    kwargs.pop("left_on", None)
    kwargs.pop("right_on", None)
    kwargs.pop("suffix", "_right")
    kwargs.pop("validate", "m:m")
    kwargs.pop("head", None)
    kwargs.pop("tail", None)
    kwargs.pop("sample", None)
    kwargs.pop("method", None)
    kwargs.pop("write_path", None)

    streaming = streaming in {"", "True", "true", "T", "t", "1"}

    if read_type is not None:
        assert read_type in DATA_TYPES, (
            str(read_type) + "not in the supported list: " + str(DATA_TYPES)
        )

    import polars as pl

    ls = []
    for path in paths:
        if read_type is None:
            _read_type = path.split(".")[-1].replace("jsonl", "ndjson")
        else:
            _read_type = read_type
        if _read_type not in DATA_TYPES:
            continue
        read_func = getattr(pl, "scan_" + _read_type, None)
        if read_func is None:
            read_func = getattr(pl, "read_" + _read_type)
            df = read_func(path, **kwargs).lazy()
        else:
            df = read_func(path, **kwargs)
        ls.append(df)

    if not ls:
        return
    elif len(ls) == 1:
        df = ls[0]
    elif join is not None:
        df = ls[0]
        for right_df in ls[1:]:
            df = df.join(
                right_df,
                on=on,
                how=join,
                left_on=left_on,
                right_on=right_on,
                suffix=suffix,
                validate=validate,
            )
    else:
        df = pl.concat(ls)

    subset_ls = []
    if head is not None:
        subset_ls.append(df.head(head))
    if tail is not None:
        subset_ls.append(df.tail(tail))
    if subset_ls:
        df = pl.concat(subset_ls)

    if sample is not None:
        df = activate(df, fetch, streaming)
        df = df.sample(sample)
        df = df.lazy()

    if method is not None:
        df = eval("df." + method)

    df = activate(df, fetch, streaming)

    if not isinstance(df, pl.DataFrame):
        text = "{}".format(df)
        if write_path:
            Path(write_path).write_text(text)
        else:
            sys.stdout.write(text)
        return

    if write_path:
        df.write_csv(write_path)
    else:
        sys.stdout.write(df.write_csv())
    return



import sys
from pathlib import Path

from .utils import get_exception_msg, subprocess_run, upath


def recur(
    kind: str,
    path: str,
    *scripts: str,
    **kwargs: str,
):
    glob = kwargs.pop("glob", "**/*")
    reverse = kwargs.pop("reverse", False)
    replace_str = kwargs.pop("replace_str", "@@")
    show_paths = kwargs.pop("show_paths", False)
    show_scripts = kwargs.pop("show_scripts", False)

    scripts = list(scripts)
    if len(kwargs) and len(scripts) == 1:
        scripts = scripts[0].split(" ")
    for k, v in kwargs.items():
        if isinstance(v, bool) and v == False:
            continue
        if len(k) >= 2:
            scripts.append("--" + k)
        elif len(k) == 1:
            scripts.append("-" + k)
        else:
            raise NotImplementedError()
        if isinstance(v, bool):
            continue
        scripts.append(str(v))

    if not scripts:
        scripts = ["echo"]

    if replace_str and all([replace_str not in script for script in scripts]):
        if len(scripts) >= 2:
            scripts.append(replace_str)
        else:
            scripts[0] += " " + replace_str

    path = Path(upath(path))
    assert path.exists(), str(path.resolve()) + " does not exist."

    if path.is_file():
        path_ls = [str(path)]
    else:
        path_ls = [str(p) for p in path.glob(glob) if p.is_file()]
        path_ls.sort(reverse=reverse)
    if show_paths:
        sys.stdout.write(
            "[Searching files]\n" + str("\n".join(["    " + p for p in path_ls]) + "\n")
        )

    running_scripts = scripts
    if kind == "under":
        for p in path_ls:
            try:
                if replace_str:
                    running_scripts = [
                        script.replace(replace_str, p) for script in scripts
                    ]
                if len(running_scripts) == 1:
                    running_scripts = running_scripts[0]
                subprocess_run(running_scripts, show_scripts)
            except Exception:
                msg = get_exception_msg()
                sys.stdout.write(msg)
                continue

    elif kind == "batch":
        if len(scripts) == 1:
            running_scripts = scripts[0]
            if replace_str:
                running_scripts = running_scripts.replace(
                    replace_str, " ".join(path_ls)
                )
        else:
            running_scripts = []
            for script in scripts:
                if replace_str and (script == replace_str):
                    running_scripts.extend(path_ls)
                else:
                    running_scripts.append(script)
        try:
            subprocess_run(running_scripts, show_scripts)
        except Exception:
            msg = get_exception_msg()
            sys.stdout.write(msg)
    else:
        raise NotImplementedError()


def under(
    path: str,
    *scripts: str,
    **kwargs: str,
):
    """Run any scripts for each file under a directory recursively."""

    kind = "under"
    recur(
        kind,
        path,
        *scripts,
        **kwargs,
    )


def batch(
    path: str,
    *scripts: str,
    **kwargs: str,
):
    """Run any scripts for a batch of files in a directory recursively."""

    kind = "batch"
    recur(
        kind,
        path,
        *scripts,
        **kwargs,
    )



import fire

from .find import find
from .pandas import pandas
from .polars import polars
from .recur import batch, under
from .search import search


def main():
    fire.Fire(
        dict(
            batch=batch,
            pandas=pandas,
            polars=polars,
            find=find,
            search=search,
            under=under,
        )
    )


if __name__ == "__main__":
    main()



import sys
import traceback
from pathlib import Path
from typing import List, Union


def get_exception_msg():
    return "".join(traceback.format_exception(*sys.exc_info()))


def upath(
    path: str,
):
    if isinstance(path, str) and path.startswith("~"):
        path = str(Path.home()) + path[1:]
    if isinstance(path, str) and path.startswith("."):
        path = str(Path.cwd()) + path[1:]
    return path


def subprocess_run(
    script: Union[str, List[str]],
    verbose: bool = True,
):
    if verbose:
        sys.stdout.write(r">>> " + script + "\n")

        from time import sleep

        sleep(0.2)

    import subprocess

    shell = isinstance(script, str)
    return subprocess.run(script, shell=shell)



import sys
from pathlib import Path


def search(
    target: str,
    path: str,
    sub: str = None,
    wildcard: str = "*",
    verbose: int = 1,
):
    """Search a keyword, which may include wildcards, in the text file content, and optionally substitute (replace)."""

    target_ls = eval("'''" + target + "'''").split(wildcard)

    p = Path(path)
    try:
        text = p.read_text()
    except Exception:
        if verbose >= 3:
            raise
        return

    replacing_ls = []
    end_index = 0

    while True:
        index = end_index
        start_index = None

        for target_ss in target_ls:
            index = text.find(target_ss, index)
            if index >= 0:
                start_index = start_index or index
                index = index + len(target_ss)
            else:
                break
        if start_index and (index >= 0):
            end_index = index
            if sub is not None:
                replacing = text[start_index:end_index]
                replacing_ls.append(replacing)
            if verbose >= 1:
                sys.stdout.write(
                    f"{p} [{start_index}:{end_index}]\n{text[start_index:end_index]}\n"
                )
        else:
            break
    for replacing in replacing_ls:
        text = text.replace(replacing, sub, 1)
    if sub is not None:
        p.write_text(text)
